/*
 * Copyright (C) Sistina Software, Inc.  1997-2003 All rights reserved.
 * Copyright (C) 2004-2006 Red Hat, Inc.  All rights reserved.
 *
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions
 * of the GNU General Public License version 2.
 */

#ifndef __GFS2_DOT_H__
#define __GFS2_DOT_H__

enum {
	NO_CREATE = 0,
	CREATE = 1,
};

enum {
	NO_WAIT = 0,
	WAIT = 1,
};

enum {
	NO_FORCE = 0,
	FORCE = 1,
};

#define GFS2_FAST_NAME_SIZE 8

#endif /* __GFS2_DOT_H__ */

