/*
 * linux/include/asm-arm/arch-ixp2000/dma.h
 *
 * Copyright (C) 2002 Intel Corp.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
