/*
 * Ours Technology Inc. OTi-6858 USB to serial adapter driver.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */
#ifndef __LINUX_USB_SERIAL_OTI6858_H
#define __LINUX_USB_SERIAL_OTI6858_H

#define OTI6858_VENDOR_ID	0x0ea0
#define OTI6858_PRODUCT_ID	0x6858

#endif
